"""
An interface to SGE.
"""

from __future__ import annotations
from collections.abc import Sequence
import re
from typing import cast, TYPE_CHECKING
from typing_extensions import override
from hpcflow.sdk.typing import hydrate
from hpcflow.sdk.core.errors import (
    IncompatibleSGEPEError,
    NoCompatibleSGEPEError,
    UnknownSGEPEError,
)
from hpcflow.sdk.log import TimeIt
from hpcflow.sdk.submission.enums import JobscriptElementState
from hpcflow.sdk.submission.schedulers import QueuedScheduler
from hpcflow.sdk.submission.schedulers.utils import run_cmd

if TYPE_CHECKING:
    from collections.abc import Iterator, Mapping
    from typing import Any, ClassVar
    from ...config.types import SchedulerConfigDescriptor
    from ...core.element import ElementResources
    from ..jobscript import Jobscript
    from ..types import VersionInfo
    from ..shells.base import Shell


@hydrate
class SGEPosix(QueuedScheduler):
    """
    A scheduler that uses SGE.

    Keyword Args
    ------------
    cwd_switch: str
        Override of default switch to use to set the current working directory.
    shell_args: str
        Arguments to pass to the shell. Pre-quoted.
    shebang_args: str
        Arguments to set on the shebang line. Pre-quoted.
    options: dict
        Options to the jobscript command.

    Notes
    -----
    - runs in serial by default

    References
    ----------
    [1] https://gridscheduler.sourceforge.net/htmlman/htmlman1/qsub.html
    [2] https://softpanorama.org/HPC/Grid_engine/Queues/queue_states.shtml

    """

    #: Default args for shebang line.
    DEFAULT_SHEBANG_ARGS: ClassVar[str] = ""
    #: Default submission command.
    DEFAULT_SUBMIT_CMD: ClassVar[str] = "qsub"
    #: Default command to show the queue state.
    DEFAULT_SHOW_CMD: ClassVar[Sequence[str]] = ("qstat",)
    #: Default cancel command.
    DEFAULT_DEL_CMD: ClassVar[str] = "qdel"
    #: Default job control directive prefix.
    DEFAULT_JS_CMD: ClassVar[str] = "#$"
    #: Default prefix to enable array processing.
    DEFAULT_ARRAY_SWITCH: ClassVar[str] = "-t"
    #: Default shell variable with array ID.
    DEFAULT_ARRAY_ITEM_VAR: ClassVar[str] = "SGE_TASK_ID"
    #: Default switch to control CWD.
    DEFAULT_CWD_SWITCH: ClassVar[str] = "-cwd"
    #: Default command to get the login nodes.
    DEFAULT_LOGIN_NODES_CMD: ClassVar[Sequence[str]] = ("qconf", "-sh")

    #: Maps scheduler state codes to :py:class:`JobscriptElementState` values.
    state_lookup: ClassVar[Mapping[str, JobscriptElementState]] = {
        "qw": JobscriptElementState.pending,
        "hq": JobscriptElementState.waiting,
        "hR": JobscriptElementState.waiting,
        "r": JobscriptElementState.running,
        "t": JobscriptElementState.running,
        "Rr": JobscriptElementState.running,
        # "Rt": JobscriptElementState.running,
        "s": JobscriptElementState.errored,
        "ts": JobscriptElementState.errored,
        "S": JobscriptElementState.errored,
        "tS": JobscriptElementState.errored,
        "T": JobscriptElementState.errored,
        "tT": JobscriptElementState.errored,
        "Rs": JobscriptElementState.errored,
        "Rt": JobscriptElementState.errored,
        "RS": JobscriptElementState.errored,
        "RT": JobscriptElementState.errored,
        "Eq": JobscriptElementState.errored,
        "Eh": JobscriptElementState.errored,
        "dr": JobscriptElementState.cancelled,
        "dt": JobscriptElementState.cancelled,
        "dR": JobscriptElementState.cancelled,
        "ds": JobscriptElementState.cancelled,
        "dS": JobscriptElementState.cancelled,
        "dT": JobscriptElementState.cancelled,
    }

    def __init__(self, cwd_switch: str | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cwd_switch = cwd_switch or self.DEFAULT_CWD_SWITCH

    @classmethod
    @override
    @TimeIt.decorator
    def process_resources(
        cls, resources: ElementResources, scheduler_config: SchedulerConfigDescriptor
    ) -> None:
        """
        Perform scheduler-specific processing to the element resources.

        Note
        ----
        This mutates `resources`.
        """
        if resources.num_nodes is not None:
            raise ValueError(
                f"Specifying `num_nodes` for the {cls.__name__!r} scheduler is not "
                f"supported."
            )

        para_envs = scheduler_config.get("parallel_environments", {})

        if resources.SGE_parallel_env is not None:
            # check user-specified `parallel_env` is valid and compatible with
            # `num_cores`:
            if resources.num_cores and resources.num_cores == 1:
                raise ValueError(
                    f"An SGE parallel environment should not be specified if `num_cores` "
                    f"is 1 (`SGE_parallel_env` was specified as "
                    f"{resources.SGE_parallel_env!r})."
                )

            try:
                env = para_envs[resources.SGE_parallel_env]
            except KeyError:
                raise UnknownSGEPEError(resources.SGE_parallel_env, para_envs)
            if not cls.is_num_cores_supported(resources.num_cores, env["num_cores"]):
                raise IncompatibleSGEPEError(
                    resources.SGE_parallel_env, resources.num_cores
                )
        else:
            # find the first compatible PE:
            for pe_name, pe_info in para_envs.items():
                if cls.is_num_cores_supported(resources.num_cores, pe_info["num_cores"]):
                    resources.SGE_parallel_env = pe_name
                    break
            else:
                raise NoCompatibleSGEPEError(resources.num_cores)

    def get_login_nodes(self) -> list[str]:
        """Return a list of hostnames of login/administrative nodes as reported by the
        scheduler."""
        get_login = self.login_nodes_cmd
        assert get_login is not None and len(get_login) >= 1
        stdout, stderr = run_cmd(get_login)
        if stderr:
            print(stderr)
        return stdout.strip().split("\n")

    def __format_core_request_lines(self, resources: ElementResources) -> Iterator[str]:
        if resources.num_cores and resources.num_cores > 1:
            yield f"{self.js_cmd} -pe {resources.SGE_parallel_env} {resources.num_cores}"
        if resources.max_array_items:
            yield f"{self.js_cmd} -tc {resources.max_array_items}"

    def __format_array_request(self, num_elements: int) -> str:
        return f"{self.js_cmd} {self.array_switch} 1-{num_elements}"

    def get_stdout_filename(
        self, js_idx: int, job_ID: str, array_idx: int | None = None
    ) -> str:
        """File name of the standard output stream file."""
        # TODO: untested, might not work!
        array_idx_str = f".{array_idx}" if array_idx is not None else ""
        return f"js_{js_idx}.sh.o{job_ID}{array_idx_str}"

    def get_stderr_filename(
        self, js_idx: int, job_ID: str, array_idx: int | None = None
    ) -> str:
        """File name of the standard error stream file."""
        # TODO: untested, might not work!
        array_idx_str = f".{array_idx}" if array_idx is not None else ""
        return f"js_{js_idx}.sh.e{job_ID}{array_idx_str}"

    def __format_std_stream_file_option_lines(
        self, is_array: bool, sub_idx: int, js_idx: int, combine_std: bool
    ) -> Iterator[str]:
        # note: if we modify the file names, there is, I believe, no way to include the
        # job ID; so we don't modify the file names:
        base = f"./artifacts/submissions/{sub_idx}/js_std/{js_idx}"
        yield f"{self.js_cmd} -o {base}"
        if combine_std:
            yield f"{self.js_cmd} -j y"  # redirect stderr to stdout
        else:
            yield f"{self.js_cmd} -e {base}"

    @override
    def format_options(
        self,
        resources: ElementResources,
        num_elements: int,
        is_array: bool,
        sub_idx: int,
        js_idx: int,
    ) -> str:
        """
        Format the options to the jobscript command.
        """
        opts: list[str] = []
        opts.append(self.format_switch(self.cwd_switch))
        opts.extend(self.__format_core_request_lines(resources))
        if is_array:
            opts.append(self.__format_array_request(num_elements))

        opts.extend(
            self.__format_std_stream_file_option_lines(
                is_array, sub_idx, js_idx, resources.combine_jobscript_std
            )
        )

        for opt_k, opt_v in self.options.items():
            if opt_v is None:
                opts.append(f"{self.js_cmd} {opt_k}")
            elif isinstance(opt_v, list):
                opts.extend(f"{self.js_cmd} {opt_k} {i}" for i in opt_v)
            elif opt_v:
                opts.append(f"{self.js_cmd} {opt_k} {opt_v}")

        return "\n".join(opts) + "\n"

    @override
    @TimeIt.decorator
    def get_version_info(self) -> VersionInfo:
        stdout, stderr = run_cmd([*self.show_cmd, "-help"])
        if stderr:
            print(stderr)
        first_line, *_ = stdout.split("\n")
        name, version, *_ = first_line.strip().split()
        return {
            "scheduler_name": name,
            "scheduler_version": version,
        }

    @override
    def get_submit_command(
        self,
        shell: Shell,
        js_path: str,
        deps: dict[Any, tuple[Any, ...]],
    ) -> list[str]:
        """
        Get the command to use to submit a job to the scheduler.

        Returns
        -------
        List of argument words.
        """
        cmd = [self.submit_cmd, "-terse"]

        dep_job_IDs: list[str] = []
        dep_job_IDs_arr: list[str] = []
        for job_ID, is_array_dep in deps.values():
            if is_array_dep:  # array dependency
                dep_job_IDs_arr.append(str(job_ID))
            else:
                dep_job_IDs.append(str(job_ID))

        if dep_job_IDs:
            cmd.append("-hold_jid")
            cmd.append(",".join(dep_job_IDs))

        if dep_job_IDs_arr:
            cmd.append("-hold_jid_ad")
            cmd.append(",".join(dep_job_IDs_arr))

        cmd.append(js_path)
        return cmd

    __SGE_JOB_ID_RE: ClassVar[re.Pattern] = re.compile(r"^\d+")

    def parse_submission_output(self, stdout: str) -> str:
        """Extract scheduler reference for a newly submitted jobscript"""
        if not (match := self.__SGE_JOB_ID_RE.search(stdout)):
            raise RuntimeError(f"Could not parse Job ID from scheduler output {stdout!r}")
        return match.group()

    def get_job_statuses(
        self,
    ) -> Mapping[str, JobscriptElementState | Mapping[int, JobscriptElementState]]:
        """Get information about all of this user's jobscripts that are currently listed
        by the scheduler."""
        cmd = [*self.show_cmd, "-u", "$USER", "-g", "d"]  # "-g d": separate arrays items
        stdout, stderr = run_cmd(cmd, logger=self._app.submission_logger)
        if stderr:
            raise ValueError(
                f"Could not get query SGE jobs. Command was: {cmd!r}; stderr was: "
                f"{stderr}"
            )
        elif not stdout:
            return {}

        info: dict[str, dict[int, JobscriptElementState] | JobscriptElementState] = {}
        lines = stdout.split("\n")
        # assuming a job name with spaces means we can't split on spaces to get
        # anywhere beyond the job name, so get the column index of the state heading
        # and assume the state is always left-aligned with the heading:
        state_idx = lines[0].index("state")
        task_id_idx = lines[0].index("ja-task-ID")
        for ln in lines[2:]:
            if not ln:
                continue
            base_job_ID, *_ = ln.split()

            # states can be one or two chars (for our limited purposes):
            state_str = ln[state_idx : state_idx + 2].strip()
            state = self.state_lookup[state_str]

            arr_idx_s = ln[task_id_idx:].strip()
            arr_idx = (
                int(arr_idx_s) - 1  # We are using zero-indexed info
                if arr_idx_s
                else None
            )

            if arr_idx is not None:
                entry = cast(
                    dict[int, JobscriptElementState], info.setdefault(base_job_ID, {})
                )
                entry[arr_idx] = state
            else:
                info[base_job_ID] = state
        return info

    @override
    def get_job_state_info(
        self, *, js_refs: Sequence[str] | None = None
    ) -> Mapping[str, JobscriptElementState | Mapping[int, JobscriptElementState]]:
        """Query the scheduler to get the states of all of this user's jobs, optionally
        filtering by specified job IDs.

        Jobs that are not in the scheduler's status output will not appear in the output
        of this method.

        """
        info = self.get_job_statuses()
        if js_refs:
            return {k: v for k, v in info.items() if k in js_refs}
        return info

    @override
    def cancel_jobs(
        self,
        js_refs: list[str],
        jobscripts: list[Jobscript] | None = None,
    ):
        """
        Cancel submitted jobs.
        """
        cmd = [self.del_cmd] + js_refs
        self._app.submission_logger.info(
            f"cancelling {self.__class__.__name__} jobscripts with command: {cmd}."
        )
        stdout, stderr = run_cmd(cmd, logger=self._app.submission_logger)
        if stderr:
            raise ValueError(
                f"Could not get query SGE {self.__class__.__name__}. Command was: "
                f"{cmd!r}; stderr was: {stderr}"
            )
        self._app.submission_logger.info(
            f"jobscripts cancel command executed; stdout was: {stdout}."
        )



import os
from textwrap import dedent

import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core.actions import EARStatus


@pytest.mark.integration
@pytest.mark.parametrize("combine_scripts", [True, False])
def test_run_status_fail_when_missing_script_output_data_file(
    null_config, tmp_path, combine_scripts
):

    s1 = hf.TaskSchema(
        objective="t1",
        outputs=[hf.SchemaOutput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                script="<<script:main_script_test_json_out_FAIL.py>>",
                script_data_out="json",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
            )
        ],
    )

    tasks = [
        hf.Task(s1),  # will fail due to not generaing an output data file
    ]

    wk = hf.Workflow.from_template_data(
        template_name="test_run_status_fail_missing_script_output_file",
        path=tmp_path,
        tasks=tasks,
        resources={
            "any": {
                "write_app_logs": True,
                "skip_downstream_on_failure": False,
                "combine_scripts": combine_scripts,
            }
        },
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    runs = wk.get_all_EARs()
    assert runs[0].status is EARStatus.error


@pytest.mark.integration
@pytest.mark.parametrize("combine_scripts", [True, False])
def test_run_status_fail_when_missing_script_output_data_file_OFP_fail(
    null_config, tmp_path, combine_scripts
):

    out_file_name = "my_output_file.txt"
    out_file = hf.FileSpec(label="my_output_file", name=out_file_name)

    if os.name == "nt":
        cmd = f"Set-Content -Path {out_file_name} -Value (<<parameter:p1>> + 100)"
    else:
        cmd = f"echo $(( <<parameter:p1>> + 100 )) > {out_file_name}"

    # this script parses the output file but then deletes this file so it can't be saved!
    act = hf.Action(
        commands=[hf.Command(cmd)],
        output_file_parsers=[
            hf.OutputFileParser(
                output_files=[out_file],
                output=hf.Parameter("p2"),
                script="<<script:output_file_parser_basic_FAIL.py>>",
                save_files=True,
            ),
        ],
        environments=[hf.ActionEnvironment(environment="python_env")],
    )

    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        outputs=[hf.SchemaInput(parameter=hf.Parameter("p2"))],
        actions=[act],
    )
    t1 = hf.Task(schema=s1, inputs={"p1": 100})

    wk = hf.Workflow.from_template_data(
        template_name="test_run_status_fail_missing_OFP_save_file",
        path=tmp_path,
        tasks=[t1],
        resources={
            "any": {
                "write_app_logs": True,
                "skip_downstream_on_failure": False,
                "combine_scripts": combine_scripts,
            }
        },
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    runs = wk.get_all_EARs()
    assert runs[0].status is EARStatus.success
    assert runs[1].status is EARStatus.error


@pytest.mark.integration
@pytest.mark.parametrize("combine_scripts", [True, False])
def test_run_status_fail_when_missing_IFG_input_file(
    null_config, tmp_path, combine_scripts
):

    inp_file = hf.FileSpec(label="my_input_file", name="my_input_file.txt")

    if os.name == "nt":
        cmd = dedent(
            """\
            try {
                Get-Content "<<file:my_input_file>>" -ErrorAction Stop
            } catch {
                Write-Host "File does not exist."
                exit 1 
            }
        """
        )
    else:
        cmd = "cat <<file:my_input_file>>"

    # this script silently fails to generate the input file!
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter=hf.Parameter("p1"))],
        actions=[
            hf.Action(
                commands=[hf.Command(cmd)],
                input_file_generators=[
                    hf.InputFileGenerator(
                        input_file=inp_file,
                        inputs=[hf.Parameter("p1")],
                        script="<<script:input_file_generator_basic_FAIL.py>>",
                    ),
                ],
                environments=[hf.ActionEnvironment(environment="python_env")],
            )
        ],
    )
    t1 = hf.Task(schema=s1, inputs={"p1": 100})
    wk = hf.Workflow.from_template_data(
        template_name="test_run_status_fail_missing_IFG_save_file",
        path=tmp_path,
        tasks=[t1],
        resources={
            "any": {
                "write_app_logs": True,
                "skip_downstream_on_failure": False,
                "combine_scripts": combine_scripts,
            }
        },
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    runs = wk.get_all_EARs()
    assert runs[0].status is EARStatus.error  # no input file to save
    assert runs[1].status is EARStatus.error  # no input file to consume


@pytest.mark.integration
@pytest.mark.parametrize("combine_scripts", [True, False])
def test_run_status_fail_when_action_save_file(null_config, tmp_path, combine_scripts):

    my_file = hf.FileSpec(label="my_file", name="my_file.txt")

    # this script does not generate a file that can be saved:
    s1 = hf.TaskSchema(
        objective="t1",
        actions=[
            hf.Action(
                script="<<script:do_nothing.py>>",
                script_exe="python_script",
                environments=[hf.ActionEnvironment(environment="python_env")],
                requires_dir=True,
                save_files=[my_file],
            )
        ],
    )
    t1 = hf.Task(schema=s1)
    wk = hf.Workflow.from_template_data(
        template_name="test_run_status_fail_missing_action_save_file",
        path=tmp_path,
        tasks=[t1],
        resources={
            "any": {
                "write_app_logs": True,
                "skip_downstream_on_failure": False,
                "combine_scripts": combine_scripts,
            }
        },
    )
    wk.submit(wait=True, add_to_known=False, status=False)
    runs = wk.get_all_EARs()
    assert runs[0].status is EARStatus.error  # no file to save



from __future__ import annotations
from dataclasses import dataclass
from pathlib import Path
import random
import string
import sys
from textwrap import dedent
from typing import ClassVar

import pytest
import requests

from hpcflow.app import app as hf
from hpcflow.sdk.core.errors import ParametersMetadataReadOnlyError
from hpcflow.sdk.typing import hydrate
from hpcflow.sdk.core.parameters import ParameterValue


@dataclass
@hydrate
class MyParameterP1(ParameterValue):
    _typ: ClassVar[str] = "p1_test"
    a: int

    def CLI_format(self):
        return str(self.a)


@pytest.mark.integration
@pytest.mark.parametrize("store", ["json", "zarr"])
def test_submission_with_specified_parameter_class_module(
    null_config, tmp_path: Path, store: str
):
    """Test we can use a ParameterValue subclass that is defined separately from the main
    code (i.e. not automatically imported on app init)."""

    # not sure why this is necessary; without this we end up with two `p1_test`
    # parameters, despite `skip_duplicates=True`:
    hf.reload_template_components()

    hf.parameters.add_object(hf.Parameter("p1_test"), skip_duplicates=True)
    s1 = hf.TaskSchema(
        objective="t1",
        inputs=[hf.SchemaInput(parameter="p1_test")],
        outputs=[hf.SchemaOutput(parameter="p2")],
        actions=[
            hf.Action(
                commands=[
                    hf.Command(
                        command="Write-Output (<<parameter:p1_test>> + 100)",
                        stdout="<<parameter:p2>>",
                    )
                ],
                rules=[
                    hf.ActionRule(
                        rule=hf.Rule(
                            path="resources.os_name",
                            condition={"value.equal_to": "nt"},
                        )
                    )
                ],
            ),
            hf.Action(
                commands=[
                    hf.Command(
                        command='echo "$((<<parameter:p1_test>> + 100))"',
                        stdout="<<parameter:p2>>",
                    )
                ],
                rules=[
                    hf.ActionRule(
                        rule=hf.Rule(
                            path="resources.os_name",
                            condition={"value.equal_to": "posix"},
                        )
                    )
                ],
            ),
        ],
        parameter_class_modules=["hpcflow.tests.unit.test_parameter"],
    )
    p1_value = MyParameterP1(a=10)
    t1 = hf.Task(schema=s1, inputs=[hf.InputValue("p1_test", value=p1_value)])
    wk = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="w1",
        path=tmp_path,
        store=store,
    )
    wk.submit(wait=True, add_to_known=False)
    assert wk.tasks.t1.elements[0].get("outputs.p2") == "110"


@pytest.mark.parametrize("store", ["json", "zarr"])
def test_unseen_parameter(null_config, tmp_path: Path, store: str):
    """Test we can generate a workflow that uses an unseen parameter type."""

    random_str = "".join(random.choice(string.ascii_letters) for _ in range(10))
    p_type = f"p_{random_str}"
    act = hf.Action(
        commands=[
            hf.Command(
                command=f'echo "$((<<parameter:{p_type}>> + 1))"',
                stdout=f"<<int(parameter:{p_type})>>",
            )
        ]
    )
    ts = hf.TaskSchema(
        objective="add_one",
        actions=[act],
        inputs=[hf.SchemaInput(p_type)],
        outputs=[hf.SchemaOutput(p_type)],
    )
    wkt = hf.WorkflowTemplate(
        name="increment_number",
        tasks=[
            hf.Task(
                schema=ts,
                inputs={p_type: 5},
            )
        ],
    )
    wk = hf.Workflow.from_template(wkt, path=tmp_path, store=store)
    assert wk.tasks[0].elements[0].get(f"inputs.{p_type}") == 5


def test_iter(new_null_config, tmp_path: Path):
    values = [1, 2, 3]
    wkt = hf.WorkflowTemplate(
        name="test",
        tasks=[
            hf.Task(
                schema=hf.task_schemas.test_t1_ps,
                sequences=[hf.ValueSequence(path="inputs.p1", values=values)],
            ),
        ],
    )
    wk = hf.Workflow.from_template(wkt, path=tmp_path)
    for idx, param_p1_i in enumerate(wk.tasks[0].inputs.p1):
        assert param_p1_i.value == values[idx]


def test_slice(new_null_config, tmp_path: Path):
    values = [1, 2, 3]
    wkt = hf.WorkflowTemplate(
        name="test",
        tasks=[
            hf.Task(
                schema=hf.task_schemas.test_t1_ps,
                sequences=[hf.ValueSequence(path="inputs.p1", values=values)],
            ),
        ],
    )
    wk = hf.Workflow.from_template(wkt, path=tmp_path)
    p1_params = wk.tasks[0].inputs.p1[0::2]
    assert len(p1_params) == 2
    assert p1_params[0].value == values[0]
    assert p1_params[1].value == values[2]


@pytest.mark.xfail(
    condition=sys.platform == "darwin",
    raises=requests.exceptions.HTTPError,
    reason=(
        "GHA MacOS runners use the same IP address, so we get rate limited when "
        "retrieving demo data from GitHub."
    ),
)
def test_demo_data_substitution_param_value_class_method(new_null_config, tmp_path: Path):
    yaml_str = dedent(
        """\
        name: temp
        template_components:
          task_schemas:
            - objective: test
              inputs:
                - parameter: p1c
              parameter_class_modules: [hpcflow.sdk.core.test_utils]
        tasks: 
          - schema: test
            inputs:
              p1c::from_file:
                path: <<demo_data_file:text_file.txt>>
    """
    )
    wk = hf.Workflow.from_YAML_string(YAML_str=yaml_str, path=tmp_path)
    assert wk.tasks[0].template.element_sets[0].inputs[0].value == {
        "path": str(hf.demo_data_cache_dir.joinpath("text_file.txt"))
    }


@pytest.mark.xfail(
    condition=sys.platform == "darwin",
    raises=requests.exceptions.HTTPError,
    reason=(
        "GHA MacOS runners use the same IP address, so we get rate limited when "
        "retrieving demo data from GitHub."
    ),
)
def test_demo_data_substitution_value_sequence_class_method(
    new_null_config, tmp_path: Path
):
    yaml_str = dedent(
        """\
        name: temp
        template_components:
          task_schemas:
            - objective: test
              inputs:
                - parameter: p1
        tasks:
          - schema: test
            sequences:
              - path: inputs.p1
                values::from_file:
                  file_path: <<demo_data_file:text_file.txt>>
    """
    )
    wk = hf.Workflow.from_YAML_string(YAML_str=yaml_str, path=tmp_path)
    assert wk.tasks[0].template.element_sets[0].sequences[0].values == [
        "1",
        "2",
        "3",
        "4",
        "5",
        "6",
        "7",
        "8",
        "9",
        "10",
    ]


def test_json_store_parameters_metadata_cache_raises_on_modify(
    null_config, tmp_path: Path
):
    wk = hf.make_demo_workflow("workflow_1", path=tmp_path, store="json")
    assert isinstance(wk, hf.Workflow)
    num_params = len(wk.get_all_parameters())
    with pytest.raises(ParametersMetadataReadOnlyError):
        with wk._store.parameters_metadata_cache():
            wk._add_unset_parameter_data(source={"type": "UNSET-FOR-THIS-TEST"})
    assert len(wk.get_all_parameters()) == num_params



from __future__ import annotations
from textwrap import dedent

import numpy as np
import pytest

from hpcflow.app import app as hf
from hpcflow.sdk.core.test_utils import make_schemas
from hpcflow.sdk.core.utils import get_file_context


def test_MPS_sequences():
    mps = hf.MultiPathSequence(paths=("inputs.p1", "inputs.p2"), values=[[0, 1], [2, 3]])
    assert mps.sequences == [
        hf.ValueSequence(path="inputs.p1", values=[0, 1]),
        hf.ValueSequence(path="inputs.p2", values=[2, 3]),
    ]


def test_MPS_sequences_moved_to_element_set():
    mps = hf.MultiPathSequence(paths=("inputs.p1", "inputs.p2"), values=[[0, 1], [2, 3]])
    es = hf.ElementSet(multi_path_sequences=[mps])
    expected_mps_seqs = [
        hf.ValueSequence(path="inputs.p1", values=[0, 1]),
        hf.ValueSequence(path="inputs.p2", values=[2, 3]),
    ]
    assert mps._sequences is None
    assert mps.sequence_indices == [0, 2]
    assert es.sequences == mps.sequences
    assert es.sequences == expected_mps_seqs


def test_MPS_sequences_moved_to_element_set_with_existing_sequences():
    mps = hf.MultiPathSequence(paths=("inputs.p1", "inputs.p2"), values=[[0, 1], [2, 3]])
    seq = hf.ValueSequence(path="inputs.p0", values=[0, 1])
    expected_mps_seqs = [
        hf.ValueSequence(path="inputs.p1", values=[0, 1]),
        hf.ValueSequence(path="inputs.p2", values=[2, 3]),
    ]
    es = hf.ElementSet(
        sequences=[seq],
        multi_path_sequences=[mps],
    )
    assert mps._sequences is None
    assert mps.sequence_indices == [1, 3]
    assert mps.sequences == expected_mps_seqs
    assert es.sequences == [seq, *expected_mps_seqs]


def test_MPS_sequences_moved_to_task_element_set():
    mps = hf.MultiPathSequence(paths=("inputs.p1", "inputs.p2"), values=[[0, 1], [2, 3]])
    s1 = make_schemas(({"p1": None, "p2": None}, ()))
    t1 = hf.Task(s1, multi_path_sequences=[mps])
    expected_mps_seqs = [
        hf.ValueSequence(path="inputs.p1", values=[0, 1]),
        hf.ValueSequence(path="inputs.p2", values=[2, 3]),
    ]
    es = t1.element_sets[0]
    assert mps._sequences is None
    assert mps.sequence_indices == [0, 2]
    assert es.sequences == mps.sequences
    assert es.sequences == expected_mps_seqs


def test_MPS_sequences_moved_to_task_element_set_with_existing_sequences():
    mps = hf.MultiPathSequence(paths=("inputs.p1", "inputs.p2"), values=[[0, 1], [2, 3]])
    seq = hf.ValueSequence(path="inputs.p0", values=[0, 1])
    s1 = make_schemas(({"p0": None, "p1": None, "p2": None}, ()))
    t1 = hf.Task(s1, sequences=[seq], multi_path_sequences=[mps])
    expected_mps_seqs = [
        hf.ValueSequence(path="inputs.p1", values=[0, 1]),
        hf.ValueSequence(path="inputs.p2", values=[2, 3]),
    ]
    es = t1.element_sets[0]
    assert mps._sequences is None
    assert mps.sequence_indices == [1, 3]
    assert mps.sequences == expected_mps_seqs
    assert es.sequences == [seq, *expected_mps_seqs]


def test_MPS_sequence_element_inputs(null_config, tmp_path):
    mps = hf.MultiPathSequence(paths=("inputs.p1", "inputs.p2"), values=[[0, 1], [2, 3]])
    s1 = make_schemas(({"p1": None, "p2": None}, ()))
    t1 = hf.Task(s1, multi_path_sequences=[mps])
    wf = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_multi_path_sequence",
        path=tmp_path,
    )
    assert len(wf.template.tasks[0].element_sets[0].sequences) == 2
    assert wf.tasks[0].num_elements == 2
    assert wf.tasks[0].elements[0].get("inputs") == {"p1": 0, "p2": 2}
    assert wf.tasks[0].elements[1].get("inputs") == {"p1": 1, "p2": 3}


def test_MPS_sequence_element_inputs_with_existing_sequence(null_config, tmp_path):
    mps = hf.MultiPathSequence(paths=("inputs.p1", "inputs.p2"), values=[[0, 1], [2, 3]])
    seq = hf.ValueSequence(path="inputs.p0", values=[0, 1])
    s1 = make_schemas(({"p0": None, "p1": None, "p2": None}, ()))
    t1 = hf.Task(s1, sequences=[seq], multi_path_sequences=[mps])
    wf = hf.Workflow.from_template_data(
        tasks=[t1],
        template_name="test_multi_path_sequence",
        path=tmp_path,
    )
    assert len(wf.template.tasks[0].element_sets[0].sequences) == 3
    assert wf.tasks[0].num_elements == 2
    assert wf.tasks[0].elements[0].get("inputs") == {"p0": 0, "p1": 0, "p2": 2}
    assert wf.tasks[0].elements[1].get("inputs") == {"p0": 1, "p1": 1, "p2": 3}

    # check the same on reload:
    wf = wf.reload()
    assert len(wf.template.tasks[0].element_sets[0].sequences) == 3
    assert wf.tasks[0].num_elements == 2
    assert wf.tasks[0].elements[0].get("inputs") == {"p0": 0, "p1": 0, "p2": 2}
    assert wf.tasks[0].elements[1].get("inputs") == {"p0": 1, "p1": 1, "p2": 3}


@pytest.mark.integration
def test_MPS_element_outputs(null_config, tmp_path):
    with get_file_context("hpcflow.tests.data", "multi_path_sequences.yaml") as file_path:
        wf = hf.make_and_submit_workflow(
            file_path,
            path=tmp_path,
            status=False,
            add_to_known=False,
            wait=True,
        )
        assert wf.tasks[0].num_elements == 2

        p2 = wf.tasks[0].elements[0].outputs.p2
        assert isinstance(p2, hf.ElementParameter)
        assert p2.value == 302

        p2 = wf.tasks[0].elements[1].outputs.p2
        assert isinstance(p2, hf.ElementParameter)
        assert p2.value == 304


def test_MPS_latin_hypercube_sequence_values():
    wft_yaml = dedent(
        """\
        name: test_latin_hypercube_sampling
        template_components:
          task_schemas:
            - objective: define_p1
              inputs:
                - parameter: p1
        tasks:
          - schema: define_p1
            inputs:
              p1: {}
            multi_path_sequences:
              - paths: [inputs.p1.a, inputs.p1.b]
                values::from_latin_hypercube:
                  num_samples: 5
    """
    )
    wft = hf.WorkflowTemplate.from_YAML_string(wft_yaml)
    es = wft.tasks[0].element_sets[0]
    assert len(es.multi_path_sequences) == 1
    mps_values = np.asarray(es.multi_path_sequences[0].values)
    assert mps_values.shape == (2, 5)
    assert len(es.sequences) == 2
    seq_1 = es.sequences[0]
    seq_2 = es.sequences[1]
    assert seq_1.path == "inputs.p1.a"
    assert seq_2.path == "inputs.p1.b"
    assert np.array_equal(np.asarray(seq_1.values), mps_values[0])
    assert np.array_equal(np.asarray(seq_2.values), mps_values[1])


def test_MPS_move_from_sequences_list():
    wft_yaml = dedent(
        """\
        name: test_latin_hypercube_sampling
        template_components:
          task_schemas:
            - objective: define_p1_p2_p3_p4
              inputs:
                - parameter: p1
                - parameter: p2
                - parameter: p3
                - parameter: p4
        tasks:
          - schema: define_p1_p2_p3_p4
            inputs:
              p1: {}
              p2: {}
              p3: {}
            
            multi_path_sequences:
              - paths: [inputs.p1.a, inputs.p1.b]
                values::from_latin_hypercube:
                  num_samples: 4             
            
            sequences:
              - paths: [inputs.p2.a, inputs.p2.b] # actually a multi-path sequence
                values::from_latin_hypercube:
                  num_samples: 4
            
              - path: inputs.p4 # a normal sequence
                values: [0, 1, 2, 3]
            
              - paths: [inputs.p3.a, inputs.p3.b] # actually a multi-path sequence
                values::from_latin_hypercube:
                  num_samples: 4
    """
    )
    wft = hf.WorkflowTemplate.from_YAML_string(wft_yaml)
    es = wft.tasks[0].element_sets[0]
    mps_lst = es.multi_path_sequences
    seq_lst = es.sequences
    assert len(mps_lst) == 3
    assert len(seq_lst) == 7  # one original plus three multi-path with two paths each

    # check ordering of multi-path sequences is preserved:
    assert mps_lst[0].paths == ["inputs.p1.a", "inputs.p1.b"]
    assert mps_lst[1].paths == ["inputs.p2.a", "inputs.p2.b"]
    assert mps_lst[2].paths == ["inputs.p3.a", "inputs.p3.b"]

    # check sensible ordering of sequences:
    assert seq_lst[0].path == "inputs.p4"
    assert seq_lst[1].path == "inputs.p1.a"
    assert seq_lst[2].path == "inputs.p1.b"
    assert seq_lst[3].path == "inputs.p2.a"
    assert seq_lst[4].path == "inputs.p2.b"
    assert seq_lst[5].path == "inputs.p3.a"
    assert seq_lst[6].path == "inputs.p3.b"



def output_file_parser_basic_FAIL(my_output_file):
    # parse the output file, but then delete it, so it cannot be saved (when run with
    # `OFP.save_files=True`):
    with my_output_file.open("r") as fp:
        out = int(fp.read().strip())
    my_output_file.unlink()
    return out



"""
Adapters for various shells.
"""
from __future__ import annotations
import os

from hpcflow.sdk.core.errors import UnsupportedShellError

from hpcflow.sdk.submission.shells.base import Shell
from hpcflow.sdk.submission.shells.bash import Bash, WSLBash
from hpcflow.sdk.submission.shells.powershell import WindowsPowerShell

#: All supported shells.
ALL_SHELLS: dict[str, dict[str, type[Shell]]] = {
    "bash": {"posix": Bash},
    "powershell": {"nt": WindowsPowerShell},
    "wsl+bash": {"nt": WSLBash},
    "wsl": {"nt": WSLBash},  # TODO: cast this to wsl+bash in ResourceSpec?
}

#: The default shell in the default config.
DEFAULT_SHELL_NAMES = {
    "posix": "bash",
    "nt": "powershell",
}


def get_supported_shells(os_name: str | None = None) -> dict[str, type[Shell]]:
    """
    Get shells supported on the current or given OS.
    """
    os_name_ = os_name or os.name
    return {k: v[os_name_] for k, v in ALL_SHELLS.items() if v.get(os_name_)}


def get_shell(shell_name: str | None, os_name: str | None = None, **kwargs) -> Shell:
    """
    Get a shell interface with the given name for a given OS (or the current one).
    """
    # TODO: apply config default shell args?

    os_name = os_name or os.name
    shell_name = (
        DEFAULT_SHELL_NAMES[os_name] if shell_name is None else shell_name.lower()
    )

    supported = get_supported_shells(os_name.lower())
    if not (shell_cls := supported.get(shell_name)):
        raise UnsupportedShellError(shell=shell_name, supported=supported)

    return shell_cls(**kwargs)



"""
Core enumeration types.
"""
from __future__ import annotations
from collections.abc import Sequence
from dataclasses import dataclass
from enum import Enum


class ActionScopeType(Enum):
    """
    Types of action scope.
    """

    #: Scope that applies to anything.
    ANY = 0
    #: Scope that only applies to main scripts.
    MAIN = 1
    #: Scope that applies to processing steps.
    PROCESSING = 2
    #: Scope that applies to input file generators.
    INPUT_FILE_GENERATOR = 3
    #: Scope that applies to output file parsers.
    OUTPUT_FILE_PARSER = 4


@dataclass(frozen=True)
class _EARStatus:
    """
    Model of the state of an EARStatus.
    """

    _value: int
    #: Symbol to use when rendering a status.
    symbol: str
    #: Colour to use when rendering a status.
    colour: str
    __doc__: str = ""


class EARStatus(_EARStatus, Enum):
    """Enumeration of all possible EAR statuses, and their associated status colour."""

    #: Not yet associated with a submission.
    pending = (
        0,
        ".",
        "grey46",
        "Not yet associated with a submission.",
    )
    #: Associated with a prepared submission that is not yet submitted.
    prepared = (
        1,
        ".",
        "grey46",
        "Associated with a prepared submission that is not yet submitted.",
    )
    #: Submitted for execution.
    submitted = (
        2,
        ".",
        "grey46",
        "Submitted for execution.",
    )
    #: Executing now.
    running = (
        3,
        "●",
        "dodger_blue1",
        "Executing now.",
    )
    #: Not attempted due to a failure of an upstream action on which this depends,
    #: or a loop termination condition being satisfied.
    skipped = (
        4,
        "s",
        "dark_orange",
        (
            "Not attempted due to a failure of an upstream action on which this depends, "
            "or a loop termination condition being satisfied."
        ),
    )
    #: Aborted by the user; downstream actions will be attempted.
    aborted = (
        5,
        "A",
        "deep_pink4",
        "Aborted by the user; downstream actions will be attempted.",
    )
    #: Probably exited successfully.
    success = (
        6,
        "■",
        "green3",
        "Probably exited successfully.",
    )
    #: Probably failed.
    error = (
        7,
        "E",
        "red3",
        "Probably failed.",
    )

    @property
    def value(self) -> int:
        #: The value of the status.
        return self._value

    @classmethod
    def get_non_running_submitted_states(cls) -> frozenset[EARStatus]:
        """Return the set of all non-running states, excluding those before submission."""
        return frozenset(
            {
                cls.skipped,
                cls.aborted,
                cls.success,
                cls.error,
            }
        )

    @property
    def rich_repr(self) -> str:
        """
        The rich representation of the value.
        """
        return f"[{self.colour}]{self.symbol}[/{self.colour}]"


class InputSourceType(Enum):
    """
    The types of input sources.
    """

    #: Input source is an import.
    IMPORT = 0
    #: Input source is local.
    LOCAL = 1
    #: Input source is a default.
    DEFAULT = 2
    #: Input source is a task.
    TASK = 3


class ParallelMode(Enum):
    """
    Potential parallel modes.
    """

    #: Use distributed-memory parallelism (e.g. MPI).
    DISTRIBUTED = 0
    #: Use shared-memory parallelism (e.g. OpenMP).
    SHARED = 1
    #: Use both distributed- and shared-memory parallelism.
    #:
    #: Note
    #: ----
    #: This is not yet implemented in any meaningful way!
    HYBRID = 2


class ParameterPropagationMode(Enum):
    """
    How a parameter is propagated.
    """

    #: Parameter is propagated implicitly.
    IMPLICIT = 0
    #: Parameter is propagated explicitly.
    EXPLICIT = 1
    #: Parameter is never propagated.
    NEVER = 2


class TaskSourceType(Enum):
    """
    The types of task-based input sources.
    """

    #: Input source is a task input.
    INPUT = 0
    #: Input source is a task output.
    OUTPUT = 1
    #: Input source is unspecified.
    ANY = 2

    @classmethod
    def names(cls) -> Sequence[str]:
        """
        Get the names of the task source types.
        """
        return cls._member_names_



"""
Types used in type-checking the persistence subsystem.
"""

from __future__ import annotations
from typing import Any, Generic, TypeVar, TYPE_CHECKING
from typing_extensions import TypedDict, NotRequired, TypeAlias

if TYPE_CHECKING:
    from collections.abc import Mapping
    from .base import StoreTask, StoreElement, StoreElementIter, StoreEAR, StoreParameter
    from ..core.json_like import JSONed
    from ..core.parameters import ParameterValue
    from ..core.types import IterableParam
    from ..typing import DataIndex, ParamSource

#: Bound type variable: :class:`StoreTask`.
AnySTask = TypeVar("AnySTask", bound="StoreTask")
#: Bound type variable: :class:`StoreElement`.
AnySElement = TypeVar("AnySElement", bound="StoreElement")
#: Bound type variable: :class:`StoreElementITer`.
AnySElementIter = TypeVar("AnySElementIter", bound="StoreElementIter")
#: Bound type variable: :class:`StoreEAR`.
AnySEAR = TypeVar("AnySEAR", bound="StoreEAR")
#: Bound type variable: :class:`StoreParameter`.
AnySParameter = TypeVar("AnySParameter", bound="StoreParameter")
#: Type of possible stored parameters.
ParameterTypes: TypeAlias = (
    "ParameterValue | list | tuple | set | dict | int | float | str | None | Any"
)


class File(TypedDict):
    """
    Descriptor for file metadata.
    """

    #: Whether to store the contents.
    store_contents: bool
    #: The path to the file.
    path: str


class FileDescriptor(TypedDict):
    """
    Descriptor for file metadata.
    """

    #: Whether this is an input file.
    is_input: bool
    #: Whether to store the contents.
    store_contents: bool
    #: Where the file will go.
    dst_path: str
    #: The path to the file.
    path: str | None
    #: Whether to delete the file after processing.
    clean_up: bool
    # The contents of the file.
    contents: NotRequired[str]


class LoopDescriptor(TypedDict):
    """
    Descriptor for loop metadata.
    """

    #: The parameters iterated over by the loop.
    iterable_parameters: dict[str, IterableParam]
    #: The parameters output by the loop, and the final task insert ID from which they
    #: are output.
    output_parameters: dict[str, int]
    #: The template data from which the loop was created.
    loop_template: NotRequired[dict[str, Any]]
    #: The number of iterations generated by a loop.
    #: Note that the type is really ``list[tuple[tuple[int, ...], int]]``
    #: but the persistence implementations don't handle tuples usefully.
    num_added_iterations: list[list[list[int] | int]]
    #: The parents of the loop.
    parents: list[str]


# TODO: This type looks familiar...
class StoreCreationInfo(TypedDict):
    """
    Information about the creation of the persistence store.
    """

    #: Information about the application.
    app_info: dict[str, Any]
    #: When the persistence store was created.
    create_time: str
    #: The unique identifier for for the store/workflow.
    id: str


class ElemMeta(TypedDict):
    """
    The kwargs supported for a StoreElement.
    """

    #: The ID of the element.
    id_: int
    #: The index of the element.
    index: int
    #: The index of the element in its element set.
    es_idx: int
    #: The indices of the element in the sequences that contain it.
    seq_idx: dict[str, int]
    #: The indices of the element's sources.
    src_idx: dict[str, int]
    #: The task associated with the element.
    task_ID: int
    #: The iteration IDs.
    iteration_IDs: list[int]


class IterMeta(TypedDict):
    """
    The kwargs supported for a StoreElementIter.
    """

    #: The index of the iteration.
    data_idx: DataIndex
    #: The EARs associated with the iteration.
    EAR_IDs: dict[int, list[int]]
    #: Whether the EARs have been initialised.
    EARs_initialised: bool
    #: The ID of the element.
    element_ID: int
    #: The loops containing the iteration.
    loop_idx: dict[str, int]
    #: The schema parameters being iterated over.
    schema_parameters: list[str]


class RunMeta(TypedDict):
    """
    The kwargs supported for StoreEAR.
    """

    #: The ID of the EAR.
    id_: int
    #: The ID of the element iteration containing the EAR.
    elem_iter_ID: int
    #: The index of the action that generated the EAR.
    action_idx: int
    #: The commands that the EAR will run.
    commands_idx: list[int]
    #: The data handled by the EAR.
    data_idx: DataIndex
    #: Metadata about the EAR.
    metadata: Metadata | None
    #: When the EAR ended, if known.
    end_time: NotRequired[str | None]
    #: The exit code of the EAR, if known.
    exit_code: int | None
    #: When the EAR started, if known.
    start_time: NotRequired[str | None]
    #: Working directory snapshot at start.
    snapshot_start: dict[str, Any] | None
    #: Working directory snapshot at end.
    snapshot_end: dict[str, Any] | None
    #: The index of the EAR in the submissions.
    submission_idx: int | None
    #: Where the EAR is set to run.
    run_hostname: str | None
    #: Whether the EAR succeeded, if known.
    success: bool | None
    #: The skip reason ID, if EAR was skipped.
    skip: int
    #: Port number used by ZeroMQ during execution.
    port_number: int | None
    #: Run ID whose command file can be used for this run (may be this run's ID).
    commands_file_ID: int | None


class TaskMeta(TypedDict):
    """
    Information about a task.
    """

    #: The ID of the task.
    id_: int
    #: The index of the task in the workflow.
    index: int
    #: The elements in the task.
    element_IDs: list[int]


class TemplateMeta(TypedDict):  # FIXME: Incomplete, see WorkflowTemplate
    """
    Metadata about a workflow template.
    """

    #: Descriptors for loops.
    loops: list[dict]
    #: Descriptors for tasks.
    tasks: list[dict]


class Metadata(TypedDict):
    """
    Workflow metadata.
    """

    #: Information about the store's creation.
    creation_info: NotRequired[StoreCreationInfo]
    #: Elements in the workflow.
    elements: NotRequired[list[ElemMeta]]
    #: Iterations in the workflow.
    iters: NotRequired[list[IterMeta]]
    #: Loops in the workflow.
    loops: NotRequired[list[LoopDescriptor]]
    #: The name of the workflow.
    name: NotRequired[str]
    #: The number of added tasks.
    num_added_tasks: NotRequired[int]
    #: The replacement workflow, if any.
    replaced_workflow: NotRequired[str]
    #: Element Action Runs in the workflow.
    runs: NotRequired[list[RunMeta]]
    #: Tasks in the workflow.
    tasks: NotRequired[list[TaskMeta]]
    #: The template that generated the workflow.
    template: NotRequired[TemplateMeta]
    #: Custom template components used.
    template_components: NotRequired[dict[str, Any]]
    #: Format for timestamps.
    ts_fmt: NotRequired[str]
    #: Format for timestamps used in naming.
    ts_name_fmt: NotRequired[str]


class TypeLookup(TypedDict, total=False):
    """
    Information for looking up the type of a parameter.

    Note
    ----
    Not a total typed dictionary.
    """

    #: Tuples involving the parameter.
    tuples: list[list[int]]
    #: Sets involving the parameter.
    sets: list[list[int]]
    #: Arrays involving the parameter.
    arrays: list[list[list[int] | int]]
    #: Masked arrays involving the parameter.
    masked_arrays: list[list[int | list[int]]]


class EncodedStoreParameter(TypedDict):
    """
    The encoding of a :class:`StoreParameter`.
    """

    #: The parameter data.
    data: Any
    #: Information for looking up the type.
    type_lookup: TypeLookup


class PersistenceCache(
    TypedDict, Generic[AnySTask, AnySElement, AnySElementIter, AnySEAR, AnySParameter]
):
    """
    Cache used internally by the persistence engine.
    """

    #: Tasks.
    tasks: dict[int, AnySTask]
    #: Elements.
    elements: dict[int, AnySElement]
    #: Element iterations.
    element_iters: dict[int, AnySElementIter]
    #: Element action runs.
    EARs: dict[int, AnySEAR]
    #: Parameter sources.
    param_sources: dict[int, ParamSource]
    #: Number of tasks.
    num_tasks: int | None
    #: Parameters.
    parameters: dict[int, AnySParameter]
    #: Number of element action runs.
    num_EARs: int | None
    #: Number of parameters.
    num_params: int | None


class ZarrAttrsDict(TypedDict):
    """
    Zarr workflow attributes descriptor.
    """

    #: Workflow name.
    name: str
    #: Timestamp format.
    ts_fmt: str
    #: Timestamp format for names.
    ts_name_fmt: str
    #: Information about the creation of the workflow and persistent store.
    creation_info: StoreCreationInfo
    #: The template used to build the workflow.
    template: TemplateMeta
    #: Custom components used to build the workflow.
    template_components: dict[str, Any]
    #: Number of tasks added.
    num_added_tasks: int
    #: Tasks in the workflow.
    tasks: list[dict[str, Any]]
    #: Loops in the workflow.
    loops: list[dict[str, Any]]
    #: Submissions by the workflow.
    submissions: list[Mapping[str, JSONed]]
    #: Replacement workflow, if any.
    replaced_workflow: NotRequired[str]



"""
Miscellaneous utilities.
"""

from __future__ import annotations
from collections import Counter
from asyncio import events
import contextvars
import contextlib
import copy
import enum
import functools
import hashlib
from itertools import accumulate, islice
from importlib import resources
import json
import keyword
import os
from pathlib import Path, PurePath
import random
import re
import socket
import string
import subprocess
from datetime import datetime, timedelta, timezone
import sys
import traceback
from typing import Literal, cast, overload, TypeVar, TYPE_CHECKING
import fsspec  # type: ignore
import numpy as np

from ruamel.yaml import YAML
from watchdog.utils.dirsnapshot import DirectorySnapshot

from hpcflow.sdk.core.errors import (
    ContainerKeyError,
    InvalidIdentifier,
    MissingVariableSubstitutionError,
)
from hpcflow.sdk.log import TimeIt
from hpcflow.sdk.utils.deferred_file import DeferredFileWriter

if TYPE_CHECKING:
    from collections.abc import Callable, Hashable, Iterable, Mapping, Sequence
    from contextlib import AbstractContextManager
    from types import ModuleType
    from typing import Any, IO, Iterator
    from typing_extensions import TypeAlias
    from numpy.typing import NDArray
    from ..typing import PathLike

T = TypeVar("T")
T2 = TypeVar("T2")
T3 = TypeVar("T3")
TList: TypeAlias = "T | list[TList]"
TD = TypeVar("TD", bound="Mapping[str, Any]")
E = TypeVar("E", bound=enum.Enum)


def make_workflow_id() -> str:
    """
    Generate a random ID for a workflow.
    """
    length = 12
    chars = string.ascii_letters + "0123456789"
    return "".join(random.choices(chars, k=length))


def get_time_stamp() -> str:
    """
    Get the current time in standard string form.
    """
    return datetime.now(timezone.utc).astimezone().strftime("%Y.%m.%d_%H:%M:%S_%z")


def get_duplicate_items(lst: Iterable[T]) -> list[T]:
    """Get a list of all items in an iterable that appear more than once, assuming items
    are hashable.

    Examples
    --------
    >>> get_duplicate_items([1, 1, 2, 3])
    [1]

    >>> get_duplicate_items([1, 2, 3])
    []

    >>> get_duplicate_items([1, 2, 3, 3, 3, 2])
    [2, 3]

    """
    return [x for x, y in Counter(lst).items() if y > 1]


def check_valid_py_identifier(name: str) -> str:
    """Check a string is (roughly) a valid Python variable identifier and return it.

    The rules are:
        1. `name` must not be empty
        2. `name` must not be a Python keyword
        3. `name` must begin with an alphabetic character, and all remaining characters
           must be alphanumeric.

    Notes
    -----
    The following attributes are passed through this function on object initialisation:
        - `ElementGroup.name`
        - `Executable.label`
        - `Parameter.typ`
        - `TaskObjective.name`
        - `TaskSchema.method`
        - `TaskSchema.implementation`
        - `Loop.name`

    """
    try:
        trial_name = name[1:].replace("_", "")  # "internal" underscores are allowed
    except TypeError:
        raise InvalidIdentifier(name) from None
    except KeyError as e:
        raise KeyError(f"unexpected name type {name}") from e
    if (
        not name
        or not (name[0].isalpha() and ((trial_name[1:] or "a").isalnum()))
        or keyword.iskeyword(name)
    ):
        raise InvalidIdentifier(name)

    return name


@overload
def group_by_dict_key_values(  # type: ignore[overload-overlap]
    lst: list[dict[T, T2]], key: T
) -> list[list[dict[T, T2]]]:
    ...


@overload
def group_by_dict_key_values(lst: list[TD], key: str) -> list[list[TD]]:
    ...


def group_by_dict_key_values(lst: list, key):
    """Group a list of dicts according to specified equivalent key-values.

    Parameters
    ----------
    lst : list of dict
        The list of dicts to group together.
    key : key value
        Dicts that have identical values for all of these keys will be grouped together
        into a sub-list.

    Returns
    -------
    grouped : list of list of dict

    Examples
    --------
    >>> group_by_dict_key_values([{'a': 1}, {'a': 2}, {'a': 1}], 'a')
    [[{'a': 1}, {'a': 1}], [{'a': 2}]]

    """

    grouped = [[lst[0]]]
    for lst_item in lst[1:]:
        for group_idx, group in enumerate(grouped):
            try:
                is_vals_equal = lst_item[key] == group[0][key]

            except KeyError:
                # dicts that do not have the `key` will be in their own group:
                is_vals_equal = False

            if is_vals_equal:
                grouped[group_idx].append(lst_item)
                break

        if not is_vals_equal:
            grouped.append([lst_item])

    return grouped


def swap_nested_dict_keys(dct: dict[T, dict[T2, T3]], inner_key: T2):
    """Return a copy where top-level keys have been swapped with a second-level inner key.

    Examples:
    ---------
    >>> swap_nested_dict_keys(
        dct={
            'p1': {'format': 'direct', 'all_iterations': True},
            'p2': {'format': 'json'},
            'p3': {'format': 'direct'},
        },
        inner_key="format",
    )
    {
        "direct": {"p1": {"all_iterations": True}, "p3": {}},
        "json": {"p2": {}},
    }

    """
    out: dict[T3, dict[T, dict[T2, T3]]] = {}
    for k, v in copy.deepcopy(dct or {}).items():
        out.setdefault(v.pop(inner_key), {})[k] = v
    return out


def _ensure_int(path_comp: Any, cur_data: Any, cast_indices: bool) -> int:
    """
    Helper for get_in_container() and set_in_container()
    """
    if isinstance(path_comp, int):
        return path_comp
    if not cast_indices:
        raise TypeError(
            f"Path component {path_comp!r} must be an integer index "
            f"since data is a sequence: {cur_data!r}."
        )
    try:
        return int(path_comp)
    except (TypeError, ValueError) as e:
        raise TypeError(
            f"Path component {path_comp!r} must be an integer index "
            f"since data is a sequence: {cur_data!r}."
        ) from e


def get_in_container(
    cont, path: Sequence, cast_indices: bool = False, allow_getattr: bool = False
):
    """
    Follow a path (sequence of indices of appropriate type) into a container to obtain
    a "leaf" value. Containers can be lists, tuples, dicts,
    or any class (with `getattr()`) if ``allow_getattr`` is True.
    """
    cur_data = cont
    err_msg = (
        "Data at path {path_comps!r} is not a sequence, but is of type "
        "{cur_data_type!r} and so sub-data cannot be extracted."
    )
    for idx, path_comp in enumerate(path):
        if isinstance(cur_data, (list, tuple)):
            cur_data = cur_data[_ensure_int(path_comp, cur_data, cast_indices)]
        elif isinstance(cur_data, dict) or hasattr(cur_data, "__getitem__"):
            try:
                cur_data = cur_data[path_comp]
            except KeyError:
                raise ContainerKeyError(path=cast("list[str]", path[: idx + 1]))
        elif allow_getattr:
            try:
                cur_data = getattr(cur_data, path_comp)
            except AttributeError:
                raise ValueError(
                    err_msg.format(cur_data_type=type(cur_data), path_comps=path[:idx])
                )
        else:
            raise ValueError(
                err_msg.format(cur_data_type=type(cur_data), path_comps=path[:idx])
            )
    return cur_data


def set_in_container(
    cont, path: Sequence, value, ensure_path=False, cast_indices=False
) -> None:
    """
    Follow a path (sequence of indices of appropriate type) into a container to update
    a "leaf" value. Containers can be lists, tuples or dicts.
    The "branch" holding the leaf to update must be modifiable.
    """
    if ensure_path:
        num_path = len(path)
        for idx in range(1, num_path):
            try:
                get_in_container(cont, path[:idx], cast_indices=cast_indices)
            except (KeyError, ValueError):
                set_in_container(
                    cont=cont,
                    path=path[:idx],
                    value={},
                    ensure_path=False,
                    cast_indices=cast_indices,
                )

    sub_data = get_in_container(cont, path[:-1], cast_indices=cast_indices)
    path_comp = path[-1]
    if isinstance(sub_data, (list, tuple)):
        path_comp = _ensure_int(path_comp, sub_data, cast_indices)
    sub_data[path_comp] = value


def get_relative_path(path1: Sequence[T], path2: Sequence[T]) -> Sequence[T]:
    """Get relative path components between two paths.

    Parameters
    ----------
    path1 : tuple of (str or int or float) of length N
    path2 : tuple of (str or int or float) of length less than or equal to N

    Returns
    -------
    relative_path : tuple of (str or int or float)
        The path components in `path1` that are not in `path2`.

    Raises
    ------
    ValueError
        If the two paths do not share a common ancestor of path components, or if `path2`
        is longer than `path1`.

    Notes
    -----
    This function behaves like a simplified `PurePath(*path1).relative_to(PurePath(*path2))`
    from the `pathlib` module, but where path components can include non-strings.

    Examples
    --------
    >>> get_relative_path(('A', 'B', 'C'), ('A',))
    ('B', 'C')

    >>> get_relative_path(('A', 'B'), ('A', 'B'))
    ()

    """

    len_path2 = len(path2)
    if len(path1) < len_path2 or any(i != j for i, j in zip(path1[:len_path2], path2)):
        raise ValueError(f"{path1!r} is not in the subpath of {path2!r}.")

    return path1[len_path2:]


def search_dir_files_by_regex(
    pattern: str | re.Pattern[str], directory: str | os.PathLike = "."
) -> list[str]:
    """Search recursively for files in a directory by a regex pattern and return matching
    file paths, relative to the given directory."""
    dir_ = Path(directory)
    return [
        str(entry.relative_to(dir_))
        for entry in dir_.rglob("*")
        if re.search(pattern, entry.name)
    ]


class PrettyPrinter:
    """
    A class that produces a nice readable version of itself with ``str()``.
    Intended to be subclassed.
    """

    def __str__(self) -> str:
        lines = [self.__class__.__name__ + ":"]
        for key, val in vars(self).items():
            lines.extend(f"{key}: {val}".split("\n"))
        return "\n    ".join(lines)


def capitalise_first_letter(chars: str) -> str:
    """
    Convert the first character of a string to upper case (if that makes sense).
    The rest of the string is unchanged.
    """
    return chars[0].upper() + chars[1:]


_STRING_VARS_RE = re.compile(r"\<\<var:(.*?)(?:\[(.*)\])?\>\>")


@TimeIt.decorator
def substitute_string_vars(string: str, variables: dict[str, str]):
    """
    Scan ``string`` and substitute sequences like ``<<var:ABC>>`` with the value
    looked up in the supplied dictionary (with ``ABC`` as the key).

    Default values for the substitution can be supplied like:
    ``<<var:ABC[default=XYZ]>>``

    Examples
    --------
    >>> substitute_string_vars("abc <<var:def>> ghi", {"def": "123"})
    "abc 123 def"
    """

    def var_repl(match_obj: re.Match[str]) -> str:
        kwargs: dict[str, str] = {}
        var_name: str = match_obj[1]
        kwargs_str: str | None = match_obj[2]
        if kwargs_str:
            for i in kwargs_str.split(","):
                k, v = i.split("=")
                kwargs[k.strip()] = v.strip()
        try:
            out = str(variables[var_name])
        except KeyError:
            if "default" in kwargs:
                out = kwargs["default"]
                print(
                    f"Using default value ({out!r}) for workflow template string "
                    f"variable {var_name!r}."
                )
            else:
                raise MissingVariableSubstitutionError(var_name, variables)
        return out

    return _STRING_VARS_RE.sub(
        repl=var_repl,
        string=string,
    )


@TimeIt.decorator
def read_YAML_str(
    yaml_str: str, typ="safe", variables: dict[str, str] | Literal[False] | None = None
) -> Any:
    """Load a YAML string. This will produce basic objects.

    Parameters
    ----------
    yaml_str:
        The YAML string to parse.
    typ:
        Load type passed to the YAML library.
    variables:
        String variables to substitute in `yaml_str`. Substitutions will be attempted if
        the file looks to contain variable references (like "<<var:name>>"). If set to
        `False`, no substitutions will occur.
    """
    if variables is not False and "<<var:" in yaml_str:
        yaml_str = substitute_string_vars(yaml_str, variables=variables or {})
    yaml = YAML(typ=typ)
    return yaml.load(yaml_str)


@TimeIt.decorator
def read_YAML_file(
    path: PathLike, typ="safe", variables: dict[str, str] | Literal[False] | None = None
) -> Any:
    """Load a YAML file. This will produce basic objects.

    Parameters
    ----------
    path:
        Path to the YAML file to parse.
    typ:
        Load type passed to the YAML library.
    variables:
        String variables to substitute in the file given by `path`. Substitutions will be
        attempted if the file looks to contain variable references (like "<<var:name>>").
        If set to `False`, no substitutions will occur.
    """
    with fsspec.open(path, "rt") as f:
        yaml_str: str = f.read()
    return read_YAML_str(yaml_str, typ=typ, variables=variables)


def write_YAML_file(obj, path: str | Path, typ: str = "safe") -> None:
    """Write a basic object to a YAML file."""
    yaml = YAML(typ=typ)
    with Path(path).open("wt") as fp:
        yaml.dump(obj, fp)


def read_JSON_string(
    json_str: str, variables: dict[str, str] | Literal[False] | None = None
) -> Any:
    """Load a JSON string. This will produce basic objects.

    Parameters
    ----------
    json_str:
        The JSON string to parse.
    variables:
        String variables to substitute in `json_str`. Substitutions will be attempted if
        the file looks to contain variable references (like "<<var:name>>"). If set to
        `False`, no substitutions will occur.
    """
    if variables is not False and "<<var:" in json_str:
        json_str = substitute_string_vars(json_str, variables=variables or {})
    return json.loads(json_str)


def read_JSON_file(path, variables: dict[str, str] | Literal[False] | None = None) -> Any:
    """Load a JSON file. This will produce basic objects.

    Parameters
    ----------
    path:
        Path to the JSON file to parse.
    variables:
        String variables to substitute in the file given by `path`. Substitutions will be
        attempted if the file looks to contain variable references (like "<<var:name>>").
        If set to `False`, no substitutions will occur.
    """
    with fsspec.open(path, "rt") as f:
        json_str: str = f.read()
    return read_JSON_string(json_str, variables=variables)


def write_JSON_file(obj, path: str | Path) -> None:
    """Write a basic object to a JSON file."""
    with Path(path).open("wt") as fp:
        json.dump(obj, fp)


def get_item_repeat_index(
    lst: Sequence[T],
    *,
    distinguish_singular: bool = False,
    item_callable: Callable[[T], Hashable] | None = None,
):
    """Get the repeat index for each item in a list.

    Parameters
    ----------
    lst : list
        Must contain hashable items, or hashable objects that are returned via `callable`
        called on each item.
    distinguish_singular : bool
        If True, items that are not repeated will have a repeat index of 0, and items that
        are repeated will have repeat indices starting from 1.
    item_callable : callable
        If specified, comparisons are made on the output of this callable on each item.

    Returns
    -------
    repeat_idx : list of int
        Repeat indices of each item (see `distinguish_singular` for details).

    """

    idx: dict[Any, list[int]] = {}
    if item_callable:
        for i_idx, item in enumerate(lst):
            idx.setdefault(item_callable(item), []).append(i_idx)
    else:
        for i_idx, item in enumerate(lst):
            idx.setdefault(item, []).append(i_idx)

    rep_idx = [0] * len(lst)
    for v in idx.values():
        start = len(v) > 1 if distinguish_singular else 0
        for i_idx, i in enumerate(v, start):
            rep_idx[i] = i_idx

    return rep_idx


def get_process_stamp() -> str:
    """
    Return a globally unique string identifying this process.

    Note
    ----
    This should only be called once per process.
    """
    return "{} {} {}".format(
        datetime.now(),
        socket.gethostname(),
        os.getpid(),
    )


_ANSI_ESCAPE_RE = re.compile(r"(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]")


def remove_ansi_escape_sequences(string: str) -> str:
    """
    Strip ANSI terminal escape codes from a string.
    """
    return _ANSI_ESCAPE_RE.sub("", string)


def get_md5_hash(obj) -> str:
    """
    Compute the MD5 hash of an object.
    This is the hash of the JSON of the object (with sorted keys) as a hex string.
    """
    json_str = json.dumps(obj, sort_keys=True)
    return hashlib.md5(json_str.encode("utf-8")).hexdigest()


def get_nested_indices(
    idx: int, size: int, nest_levels: int, raise_on_rollover: bool = False
) -> list[int]:
    """Generate the set of nested indices of length `n` that correspond to a global
    `idx`.

    Examples
    --------
    >>> for i in range(4**2): print(get_nest_index(i, nest_levels=2, size=4))
    [0, 0]
    [0, 1]
    [0, 2]
    [0, 3]
    [1, 0]
    [1, 1]
    [1, 2]
    [1, 3]
    [2, 0]
    [2, 1]
    [2, 2]
    [2, 3]
    [3, 0]
    [3, 1]
    [3, 2]
    [3, 3]

    >>> for i in range(4**3): print(get_nested_indices(i, nest_levels=3, size=4))
    [0, 0, 0]
    [0, 0, 1]
    [0, 0, 2]
    [0, 0, 3]
    [0, 1, 0]
       ...
    [3, 2, 3]
    [3, 3, 0]
    [3, 3, 1]
    [3, 3, 2]
    [3, 3, 3]
    """
    if raise_on_rollover and idx >= size**nest_levels:
        raise ValueError(
            f"`idx` ({idx}) is greater than or equal to  size**nest_levels` "
            f"({size**nest_levels})."
        )

    return [(idx // (size ** (nest_levels - (i + 1)))) % size for i in range(nest_levels)]


def ensure_in(item: T, lst: list[T]) -> int:
    """Get the index of an item in a list and append the item if it is not in the
    list."""
    # TODO: add tests
    try:
        return lst.index(item)
    except ValueError:
        lst.append(item)
        return len(lst) - 1


def list_to_dict(
    lst: Sequence[Mapping[T, T2]], exclude: Iterable[T] | None = None
) -> dict[T, list[T2]]:
    """
    Convert a list of dicts to a dict of lists.
    """
    # TODO: test
    exc = frozenset(exclude or ())
    dct: dict[T, list[T2]] = {k: [] for k in lst[0] if k not in exc}
    for d in lst:
        for k, v in d.items():
            if k not in exc:
                dct[k].append(v)
    return dct


def bisect_slice(selection: slice, len_A: int) -> tuple[slice, slice]:
    """Given two sequences (the first of which of known length), get the two slices that
    are equivalent to a given slice if the two sequences were combined."""

    if selection.start < 0 or selection.stop < 0 or selection.step < 0:
        raise NotImplementedError("Can't do negative slices yet.")

    A_idx = selection.indices(len_A)
    B_start = selection.start - len_A
    if len_A != 0 and B_start < 0:
        B_start = B_start % selection.step
    if len_A > selection.stop:
        B_stop = B_start
    else:
        B_stop = selection.stop - len_A

    return slice(*A_idx), slice(B_start, B_stop, selection.step)


def replace_items(lst: list[T], start: int, end: int, repl: list[T]) -> list[T]:
    """Replaced a range of items in a list with items in another list."""
    # Convert to actual indices for our safety checks; handles end-relative addressing
    real_start, real_end, _ = slice(start, end).indices(len(lst))
    if real_end <= real_start:
        raise ValueError(
            f"`end` ({end}) must be greater than or equal to `start` ({start})."
        )
    if real_start >= len(lst):
        raise ValueError(f"`start` ({start}) must be less than length ({len(lst)}).")
    if real_end > len(lst):
        raise ValueError(
            f"`end` ({end}) must be less than or equal to length ({len(lst)})."
        )

    lst = list(lst)
    lst[start:end] = repl
    return lst


def flatten(
    lst: list[int] | list[list[int]] | list[list[list[int]]],
) -> tuple[list[int], tuple[list[int], ...]]:
    """Flatten an arbitrarily (but of uniform depth) nested list and return shape
    information to enable un-flattening.

    Un-flattening can be performed with the :py:func:`reshape` function.

    lst
        List to be flattened. Each element must contain all lists or otherwise all items
        that are considered to be at the "bottom" of the nested structure (e.g. integers).
        For example, `[[1, 2], [3]]` is permitted and flattens to `[1, 2, 3]`, but
        `[[1, 2], 3]` is not permitted because the first element is a list, but the second
        is not.

    """

    def _flatten(
        lst: list[int] | list[list[int]] | list[list[list[int]]], depth=0
    ) -> list[int]:
        out: list[int] = []
        for item in lst:
            if isinstance(item, list):
                out.extend(_flatten(item, depth + 1))
                all_lens[depth].append(len(item))
            else:
                out.append(item)
        return out

    def _get_max_depth(lst: list[int] | list[list[int]] | list[list[list[int]]]) -> int:
        val: Any = lst
        max_depth = 0
        while isinstance(val, list):
            max_depth += 1
            try:
                val = val[0]
            except IndexError:
                # empty list, assume this is max depth
                break
        return max_depth

    max_depth = _get_max_depth(lst) - 1
    all_lens: tuple[list[int], ...] = tuple([] for _ in range(max_depth))

    return _flatten(lst), all_lens


def reshape(lst: Sequence[T], lens: Sequence[Sequence[int]]) -> list[TList[T]]:
    """
    Reverse the destructuring of the :py:func:`flatten` function.
    """

    def _reshape(lst: list[T2], lens: Sequence[int]) -> list[list[T2]]:
        lens_acc = [0, *accumulate(lens)]
        return [lst[lens_acc[idx] : lens_acc[idx + 1]] for idx in range(len(lens))]

    result: list[TList[T]] = list(lst)
    for lens_i in lens[::-1]:
        result = cast("list[TList[T]]", _reshape(result, lens_i))

    return result


@overload
def remap(
    lst: list[int], mapping_func: Callable[[Sequence[int]], Sequence[T]]
) -> list[T]:
    ...


@overload
def remap(
    lst: list[list[int]], mapping_func: Callable[[Sequence[int]], Sequence[T]]
) -> list[list[T]]:
    ...


@overload
def remap(
    lst: list[list[list[int]]], mapping_func: Callable[[Sequence[int]], Sequence[T]]
) -> list[list[list[T]]]:
    ...


def remap(lst, mapping_func):
    """
    Apply a mapping to a structure of lists with ints (typically indices) as leaves to
    get a structure of lists with some objects as leaves.

    Parameters
    ----------
    lst: list[int] | list[list[int]] | list[list[list[int]]]
        The structure to remap.
    mapping_func: Callable[[Sequence[int]], Sequence[T]]
        The mapping function from sequences of ints to sequences of objects.

    Returns
    -------
    list[T] | list[list[T]] | list[list[list[T]]]
        Nested list structure in same form as input, with leaves remapped.
    """
    x, y = flatten(lst)
    return reshape(mapping_func(x), y)


_FSSPEC_URL_RE = re.compile(r"(?:[a-z0-9]+:{1,2})+\/\/")


def is_fsspec_url(url: str) -> bool:
    """
    Test if a URL appears to be one that can be understood by fsspec.
    """
    return bool(_FSSPEC_URL_RE.match(url))


class JSONLikeDirSnapShot(DirectorySnapshot):
    """
    Overridden DirectorySnapshot from watchdog to allow saving and loading from JSON.

    Parameters
    ----------
    root_path: str
        Where to take the snapshot based at.
    data: dict[str, list]
        Serialised snapshot to reload from.
        See :py:meth:`to_json_like`.
    """

    def __init__(
        self,
        root_path: str | None = None,
        data: dict[str, list] | None = None,
        use_strings: bool = False,
    ):
        """
        Create an empty snapshot or load from JSON-like data.
        """

        #: Where to take the snapshot based at.
        self.root_path = root_path
        self._stat_info: dict[str, os.stat_result] = {}
        self._inode_to_path: dict[tuple, str] = {}

        if data:
            assert root_path
            for name, item in data.items():
                # add root path
                full_name = str(PurePath(root_path) / PurePath(name))
                item = [int(i) for i in item] if use_strings else item
                stat_dat, inode_key = item[:-2], item[-2:]
                self._stat_info[full_name] = os.stat_result(stat_dat)
                self._inode_to_path[tuple(inode_key)] = full_name

    def take(self, *args, **kwargs) -> None:
        """Take the snapshot."""
        super().__init__(*args, **kwargs)

    def to_json_like(self, use_strings: bool = False) -> dict[str, Any]:
        """Export to a dict that is JSON-compatible and can be later reloaded.

        The last two integers in `data` for each path are the keys in
        `self._inode_to_path`.

        """
        # first key is the root path:
        root_path = next(iter(self._stat_info))

        # store efficiently:
        inode_invert = {v: k for k, v in self._inode_to_path.items()}
        data: dict[str, list] = {
            str(PurePath(k).relative_to(root_path)): [
                str(i) if use_strings else i for i in [*v, *inode_invert[k]]
            ]
            for k, v in self._stat_info.items()
        }

        return {
            "root_path": root_path,
            "data": data,
            "use_strings": use_strings,
        }


def open_file(filename: str | Path):
    """Open a file or directory using the default system application."""
    if sys.platform == "win32":
        os.startfile(filename)
    else:
        opener = "open" if sys.platform == "darwin" else "xdg-open"
        subprocess.call([opener, filename])


@overload
def get_enum_by_name_or_val(enum_cls: type[E], key: None) -> None:
    ...


@overload
def get_enum_by_name_or_val(enum_cls: type[E], key: str | int | float | E) -> E:
    ...


def get_enum_by_name_or_val(
    enum_cls: type[E], key: str | int | float | E | None
) -> E | None:
    """Retrieve an enum by name or value, assuming uppercase names and integer values."""
    if key is None or isinstance(key, enum_cls):
        return key
    elif isinstance(key, (int, float)):
        return enum_cls(int(key))  # retrieve by value
    elif isinstance(key, str):
        try:
            return cast("E", getattr(enum_cls, key.upper()))  # retrieve by name
        except AttributeError:
            pass
    raise ValueError(f"Unknown enum key or value {key!r} for class {enum_cls!r}")


_PARAM_SPLIT_RE = re.compile(r"((?:\w|\.)+)(?:\[(\w+)\])?")


def split_param_label(param_path: str) -> tuple[str, str] | tuple[None, None]:
    """Split a parameter path into the path and the label, if present."""
    if match := _PARAM_SPLIT_RE.match(param_path):
        return match[1], match[2]
    else:
        return None, None


def process_string_nodes(data: T, str_processor: Callable[[str], str]) -> T:
    """Walk through a nested data structure and process string nodes using a provided
    callable."""

    if isinstance(data, dict):
        return cast(
            "T", {k: process_string_nodes(v, str_processor) for k, v in data.items()}
        )

    elif isinstance(data, (list, tuple, set, frozenset)):
        _data = (process_string_nodes(i, str_processor) for i in data)
        if isinstance(data, tuple):
            return cast("T", tuple(_data))
        elif isinstance(data, set):
            return cast("T", set(_data))
        elif isinstance(data, frozenset):
            return cast("T", frozenset(_data))
        else:
            return cast("T", list(_data))

    elif isinstance(data, str):
        return cast("T", str_processor(data))

    return data


def linspace_rect(
    start: Sequence[float],
    stop: Sequence[float],
    num: Sequence[int],
    include: Sequence[str] | None = None,
    **kwargs,
) -> NDArray:
    """Generate a linear space around a rectangle.

    Parameters
    ----------
    start
        Two start values; one for each dimension of the rectangle.
    stop
        Two stop values; one for each dimension of the rectangle.
    num
        Two number values; one for each dimension of the rectangle.
    include
        If specified, include only the specified edges. Choose from "top", "right",
        "bottom", "left".

    Returns
    -------
    rect
        Coordinates of the rectangle perimeter.

    """

    if num[0] <= 1 or num[1] <= 1:
        raise ValueError("Both values in `num` must be greater than 1.")

    inc = set(include) if include else {"top", "right", "bottom", "left"}

    c0_range = np.linspace(start=start[0], stop=stop[0], num=num[0], **kwargs)
    c1_range_all = np.linspace(start=start[1], stop=stop[1], num=num[1], **kwargs)

    c1_range = c1_range_all
    if "bottom" in inc:
        c1_range = c1_range[1:]
    if "top" in inc:
        c1_range = c1_range[:-1]

    c0_range_c1_start = np.vstack([c0_range, np.repeat(start[1], num[0])])
    c0_range_c1_stop = np.vstack([c0_range, np.repeat(c1_range_all[-1], num[0])])

    c1_range_c0_start = np.vstack([np.repeat(start[0], len(c1_range)), c1_range])
    c1_range_c0_stop = np.vstack([np.repeat(c0_range[-1], len(c1_range)), c1_range])

    stacked = []
    if "top" in inc:
        stacked.append(c0_range_c1_stop)
    if "right" in inc:
        stacked.append(c1_range_c0_stop)
    if "bottom" in inc:
        stacked.append(c0_range_c1_start)
    if "left" in inc:
        stacked.append(c1_range_c0_start)

    return np.hstack(stacked)


def dict_values_process_flat(
    d: Mapping[T, T2 | list[T2]], callable: Callable[[list[T2]], list[T3]]
) -> Mapping[T, T3 | list[T3]]:
    """
    Return a copy of a dict, where the values are processed by a callable that is to
    be called only once, and where the values may be single items or lists of items.

    Examples
    --------
    d = {'a': 0, 'b': [1, 2], 'c': 5}
    >>> dict_values_process_flat(d, callable=lambda x: [i + 1 for i in x])
    {'a': 1, 'b': [2, 3], 'c': 6}

    """
    flat: list[T2] = []  # values of `d`, flattened
    is_multi: list[
        tuple[bool, int]
    ] = []  # whether a list, and the number of items to process
    for i in d.values():
        if isinstance(i, list):
            flat.extend(cast("list[T2]", i))
            is_multi.append((True, len(i)))
        else:
            flat.append(cast("T2", i))
            is_multi.append((False, 1))

    processed = callable(flat)

    out: dict[T, T3 | list[T3]] = {}
    for idx_i, (m, k) in enumerate(zip(is_multi, d)):
        start_idx = sum(i[1] for i in is_multi[:idx_i])
        end_idx = start_idx + m[1]
        proc_idx_k = processed[start_idx:end_idx]
        if not m[0]:
            out[k] = proc_idx_k[0]
        else:
            out[k] = proc_idx_k

    return out


def nth_key(dct: Iterable[T], n: int) -> T:
    """
    Given a dict in some order, get the n'th key of that dict.
    """
    it = iter(dct)
    next(islice(it, n, n), None)
    return next(it)


def nth_value(dct: dict[Any, T], n: int) -> T:
    """
    Given a dict in some order, get the n'th value of that dict.
    """
    return dct[nth_key(dct, n)]


def normalise_timestamp(timestamp: datetime) -> datetime:
    """
    Force a timestamp to have UTC as its timezone,
    then convert to use the local timezone.
    """
    return timestamp.replace(tzinfo=timezone.utc).astimezone()


def parse_timestamp(timestamp: str | datetime, ts_fmt: str) -> datetime:
    """
    Standard timestamp parsing.
    Ensures that timestamps are internally all UTC.
    """
    return normalise_timestamp(
        timestamp
        if isinstance(timestamp, datetime)
        else datetime.strptime(timestamp, ts_fmt)
    )


def current_timestamp() -> datetime:
    """
    Get a UTC timestamp for the current time
    """
    return datetime.now(timezone.utc)


def timedelta_format(td: timedelta) -> str:
    """
    Convert time delta to string in standard form.
    """
    days, seconds = td.days, td.seconds
    hours = seconds // (60 * 60)
    seconds -= hours * (60 * 60)
    minutes = seconds // 60
    seconds -= minutes * 60
    return f"{days}-{hours:02}:{minutes:02}:{seconds:02}"


_TD_RE = re.compile(r"(\d+)-(\d+):(\d+):(\d+)")


def timedelta_parse(td_str: str) -> timedelta:
    """
    Parse a string in standard form as a time delta.
    """
    if not (m := _TD_RE.fullmatch(td_str)):
        raise ValueError("not a supported timedelta form")
    days, hours, mins, secs = map(int, m.groups())
    return timedelta(days=days, hours=hours, minutes=mins, seconds=secs)


def open_text_resource(package: ModuleType | str, resource: str) -> IO[str]:
    """
    Open a file in a package.
    """
    try:
        return resources.files(package).joinpath(resource).open("r")
    except AttributeError:
        # < python 3.9; `resource.open_text` deprecated since 3.11
        return resources.open_text(package, resource)


def get_file_context(
    package: ModuleType | str, src: str | None = None
) -> AbstractContextManager[Path]:
    """
    Find a file or directory in a package.
    """
    try:
        files = resources.files(package)
        return resources.as_file(files.joinpath(src) if src else files)
        # raises ModuleNotFoundError
    except AttributeError:
        # < python 3.9
        return resources.path(package, src or "")


@contextlib.contextmanager
def redirect_std_to_file(
    file,
    mode: Literal["w", "a"] = "a",
    ignore: Callable[[BaseException], Literal[True] | int] | None = None,
) -> Iterator[None]:
    """Temporarily redirect both stdout and stderr to a file, and if an exception is
    raised, catch it, print the traceback to that file, and exit.

    File creation is deferred until an actual write is required.

    Parameters
    ----------
    ignore
        Callable to test if a given exception should be ignored. If an exception is
        not ignored, its traceback will be printed to `file` and the program will
        exit with exit code 1. The callable should accept one parameter, the
        exception, and should return True if that exception should be ignored, or
        an integer representing the exit code to exit the program with if that
        exception should not be ignored.  By default, no exceptions are ignored.

    """
    ignore = ignore or (lambda _: 1)
    with DeferredFileWriter(file, mode=mode) as fp:
        with contextlib.redirect_stdout(fp):
            with contextlib.redirect_stderr(fp):
                try:
                    yield
                except BaseException as exc:
                    ignore_ret = ignore(exc)
                    if ignore_ret is not True:
                        traceback.print_exc()
                        sys.exit(ignore_ret)


async def to_thread(func, /, *args, **kwargs):
    """Copied from https://github.com/python/cpython/blob/4b4227b907a262446b9d276c274feda2590a4e6e/Lib/asyncio/threads.py
    to support Python 3.8, which does not have `asyncio.to_thread`.

    Asynchronously run function *func* in a separate thread.

    Any *args and **kwargs supplied for this function are directly passed
    to *func*. Also, the current :class:`contextvars.Context` is propagated,
    allowing context variables from the main thread to be accessed in the
    separate thread.

    Return a coroutine that can be awaited to get the eventual result of *func*.
    """
    loop = events.get_running_loop()
    ctx = contextvars.copy_context()
    func_call = functools.partial(ctx.run, func, *args, **kwargs)
    return await loop.run_in_executor(None, func_call)



from hpcflow.sdk.utils.deferred_file import DeferredFileWriter


def test_file_not_created(tmp_path):
    file_name = tmp_path / "test.txt"
    assert not file_name.is_file()
    with DeferredFileWriter(file_name, mode="w") as fp:
        assert not fp._is_open
    assert not file_name.is_file()


def test_append_file_not_opened(tmp_path):
    file_name = tmp_path / "test.txt"
    with DeferredFileWriter(file_name, mode="a") as fp:
        assert not fp._is_open
    assert not file_name.is_file()


def test_file_created_write(tmp_path):
    file_name = tmp_path / "test.txt"
    assert not file_name.is_file()
    with DeferredFileWriter(file_name, mode="w") as fp:
        fp.write("contents\n")
        assert fp._is_open
    assert file_name.is_file()


def test_file_created_writelines(tmp_path):
    file_name = tmp_path / "test.txt"
    assert not file_name.is_file()
    with DeferredFileWriter(file_name, mode="w") as fp:
        fp.writelines(["contents\n"])
        assert fp._is_open
    assert file_name.is_file()



import h5py  # type: ignore


def main_script_test_hdf5_in_obj_2(p2, _input_files):
    # read inputs
    with h5py.File(_input_files["hdf5"], mode="r") as fh:
        a = fh["p1c"].attrs["a"].item()

    # process
    p3 = a + 100

    return {"p3": p3}



from typing import Iterable


def shorten_list_str(
    lst: Iterable, items: int = 10, end_num: int = 1, placeholder: str = "..."
) -> str:
    """Format a list as a string, including only some maximum number of items.

    Parameters
    ----------
    lst:
        The list to format in a shortened form.
    items:
        The total number of items to include in the formatted list.
    end_num:
        The number of items to include at the end of the formatted list.
    placeholder
        The placeholder to use to replace excess items in the formatted list.

    Examples
    --------
    >>> shorten_list_str(list(range(20)), items=5)
    '[0, 1, 2, 3, ..., 19]'

    """
    lst = list(lst)
    if len(lst) <= items + 1:  # (don't replace only one item)
        lst_short = lst
    else:
        start_num = items - end_num
        lst_short = lst[:start_num] + ["..."] + lst[-end_num:]

    return "[" + ", ".join(f"{i}" for i in lst_short) + "]"



import pytest

from click.testing import CliRunner
import click.exceptions

from hpcflow import __version__
from hpcflow.app import app as hf
from hpcflow.sdk.cli import ErrorPropagatingClickContext
from hpcflow.sdk.cli_common import BoolOrString


def test_version() -> None:
    runner = CliRunner()
    result = runner.invoke(hf.cli, args="--version")
    assert result.output.strip() == f"hpcFlow, version {__version__}"


def test_BoolOrString_convert():
    param_type = BoolOrString(["a"])
    assert param_type.convert(True, None, None) == True
    assert param_type.convert(False, None, None) == False
    assert param_type.convert("yes", None, None) == True
    assert param_type.convert("no", None, None) == False
    assert param_type.convert("on", None, None) == True
    assert param_type.convert("off", None, None) == False
    assert param_type.convert("a", None, None) == "a"
    with pytest.raises(click.exceptions.BadParameter):
        param_type.convert("b", None, None)


def test_error_propagated_with_custom_context_class():
    class MyException(ValueError):
        pass

    class MyContextManager:

        # set to True when MyException is raised within this context manager
        raised = False

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            if exc_type == MyException:
                self.__class__.raised = True

    @click.group()
    @click.pass_context
    def cli(ctx):
        ctx.with_resource(MyContextManager())

    cli.context_class = ErrorPropagatingClickContext  # use custom click Context

    @cli.command()
    def my_command():
        raise MyException()

    runner = CliRunner()
    runner.invoke(cli, args="my-command")

    assert MyContextManager.raised


def test_error_not_propagated_without_custom_context_class():
    class MyException(ValueError):
        pass

    class MyContextManager:

        # set to True when MyException is raised within this context manager
        raised = False

        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            if exc_type == MyException:
                self.__class__.raised = True

    @click.group()
    @click.pass_context
    def cli(ctx):
        ctx.with_resource(MyContextManager())

    @cli.command()
    def my_command():
        raise MyException()

    runner = CliRunner()
    runner.invoke(cli, args="my-command")

    assert not MyContextManager.raised


def test_std_stream_file_created(tmp_path):
    """Test exception is intercepted and printed to the specified --std-stream file."""
    error_file = tmp_path / "std_stream.txt"
    runner = CliRunner()
    result = runner.invoke(
        hf.cli, args=f'--std-stream "{str(error_file)}" internal noop --raise'
    )
    assert error_file.is_file()
    std_stream_contents = error_file.read_text()
    assert "ValueError: internal noop raised!" in std_stream_contents
    assert result.exit_code == 1
    assert result.exc_info[0] == SystemExit


def test_std_stream_file_not_created(null_config, tmp_path):
    """Test std stream file is not created when no ouput/errors/exceptions"""
    error_file = tmp_path / "std_stream.txt"
    runner = CliRunner()
    result = runner.invoke(hf.cli, args=f'--std-stream "{str(error_file)}" internal noop')
    assert not error_file.is_file()
    assert result.exit_code == 0


def test_cli_exception():
    """Test exception is passed to click"""
    runner = CliRunner()
    result = runner.invoke(hf.cli, args="internal noop --raise")
    assert result.exit_code == 1
    assert result.exc_info[0] == ValueError


def test_cli_click_exit_code_zero(tmp_path):
    """Test Click's `Exit` exception is ignored by the `redirect_std_to_file` context manager when the exit code is zero."""
    error_file = tmp_path / "std_stream.txt"
    runner = CliRunner()
    result = runner.invoke(
        hf.cli, args=f'--std-stream "{str(error_file)}" internal noop --click-exit-code 0'
    )
    assert result.exit_code == 0
    assert not error_file.is_file()


def test_cli_click_exit_code_non_zero(tmp_path):
    """Test Click's `Exit` exception is not ignored by the `redirect_std_to_file` context manager when the exit code is non-zero."""
    error_file = tmp_path / "std_stream.txt"
    runner = CliRunner()
    result = runner.invoke(
        hf.cli, args=f'--std-stream "{str(error_file)}" internal noop --click-exit-code 2'
    )
    assert result.exit_code == 2
    assert error_file.is_file()
